import gym
from gym import spaces

import numpy as np
from abc import ABC, abstractmethod, abstractproperty
import pickle
import pygame
from . import adversarial

class TicTacToeActionSpace(adversarial.AdversarialActionSpace):

    def __init__(self, env):
        self.env = env
        
    @property
    def legal_actions(self):
        """
        Returns:
            legal_actions: Returns a list of all the legal moves in the current position.
        """
        actions = []

        # Get all the empty squares (color == 0)
        s = self.env.size
        for x in range(s):
            for y in range(s):
                if self.env.board[x][y] == 0:
                    raveled_ind = np.ravel_multi_index((x,y), (s, s))
                    actions.append(raveled_ind)
        return actions
    
    @property
    def action_space_size(self):
        s = self.env.size
        return s * s

  
class TicTacToeEnv(adversarial.AdversarialEnv):
    """Abstract TicTacToe Environment"""
    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 4}

    def __init__(self, render_mode=None, render_size=512, size=3):
        "Set up initial board configuration."

        self.player_X = 1
        self.player_O = -1
        self.draw = 0
        
        self.size = size
        self.render_size = render_size
        self.reset()

        self.action_space = TicTacToeActionSpace(self)
        self.observation_space = spaces.Tuple(spaces=(
            spaces.Box(low=self.player_O, high=self.player_X, shape=(self.size, self.size), dtype=np.int8),
            spaces.Box(low=np.array([self.player_O]),
                       high=np.array([self.player_X]), dtype=np.int8)
        ))

        assert render_mode is None or render_mode in self.metadata["render_modes"]
        self.render_mode = render_mode

        self.clock = None
        self.window = None

    @property
    def current_player(self):
        """
        Returns:
            current_player: Returns identifyier for which player current has their turn.
        """        
        return self._current_player

    @property
    def previous_player(self):
        """
        Returns:
            previous_player: Returns identifyier for which player previously has their turn.
        """
        return -self.current_player

    @property
    def starting_player(self):
        return self.player_X

    def get_string_representation(self):
        """
        Returns:
            boardString: Returns string representation of current game state.
        """
        # return self.board.tobytes().hex() + f"#{self.size}"
        return pickle.dumps([self.board, self._current_player, self.size])
    
    def set_string_representation(self, board_string):
        """
        Input:
            boardString: sets game state to match the string representation of board_string.
        """
        # board, size = board_string.split('#')
        # self.size = int(size)
        # self.board = np.frombuffer(bytes.fromhex(board), dtype=self.board.dtype).reshape((self.size, self.size))
        # player = np.sum(self.board)
        # self._current_player = self.player_X if player==0 else self.player_O
        # self.board.setflags(write=True)
        # self.action_space = TicTacToeActionSpace(self)
        self.board, self._current_player, self.size = pickle.loads(board_string)
        return self._get_canonical_observation()

    def _get_canonical_observation(self):
        """
        Returns:
            canonicalState: returns canonical form of board. The canonical form
                            should be independent of players turn. For e.g. in chess,
                            the canonical form can be chosen to be from the pov
                            of white. When the player is white, we can return
                            board as is. When the player is black, we can invert
                            the colors and return the board.
        """
        return self.board * self.current_player, np.array([self.current_player], dtype=np.int8)

    def _get_info(self):
        return {}

    def _game_result(self):
        """
        Returns:
            winner: returns None when game is not finished else returns int value 
                    for the winning player or draw.
               
        """

        for row in self.board:
            if (row == row[0]).all() and row[0] != 0:
                result = row[0]
                reward = 1
                return result, reward

        for column in self.board.T:
            if  (column == column[0]).all() and column[0] != 0:
                result = column[0]
                reward = 1
                return result, reward

        for diagonal in [np.diag(self.board), np.diag(self.board[:, ::-1])]:
            if (diagonal == diagonal[0]).all() and diagonal[0] != 0:
                result = diagonal[0]
                reward = 1
                return result, reward

        # check that the game is complete. If not return None
        if 0 in self.board:
            return None, 0
        
        return self.draw, 1e-4

    def _do_action(self, action):
        s = self.size
        unraveled_action = np.unravel_index(action, (s, s))
        # Add the piece to the empty square.
        assert self.board[unraveled_action] == 0
        self.board[unraveled_action] = self.current_player
        self._current_player = self.previous_player
    
    def _reset_game(self):
        self.board = np.zeros((self.size, self.size), dtype=np.int8)
        self._current_player = self.player_X

    def _get_frame(self):

        canvas = pygame.Surface((self.render_size, self.render_size))
        BG = (210, 180, 140)
        CR = (255, 204, 203)
        CI = (144, 238, 144)
        LI = (35, 31, 32)
        canvas.fill(BG)
        pix_square_size = (
            self.render_size / self.size
        )  # The size of a single grid square in pixels
        LW = int(pix_square_size / 15)
        for x in range(self.size + 1):
            pygame.draw.line(
                canvas,
                LI,
                (0, pix_square_size * x),
                (self.render_size, pix_square_size * x),
                width=LW,
            )
            pygame.draw.line(
                canvas,
                LI,
                (pix_square_size * x, 0),
                (pix_square_size * x, self.render_size),
                width=LW,
            )
        for x in range(self.size):
            for y in range(self.size):
                piece = self.board[x, y]
                center = (pix_square_size * (0.5 + x), pix_square_size * (0.5 + y))
                if piece == self.player_X: 
                    pygame.draw.line( 
                        canvas, 
                        CR, 
                        tuple(np.add(center,(pix_square_size/3, pix_square_size/3))), # start
                        tuple(np.add(center,(-pix_square_size/3, -pix_square_size/3))), # end
                        LW 
                    )
                    pygame.draw.line( 
                        canvas, 
                        CR, 
                        tuple(np.add(center,(-pix_square_size/3, pix_square_size/3))), # start
                        tuple(np.add(center,(pix_square_size/3, -pix_square_size/3))), # end
                        LW
                    )
                elif piece == self.player_O: 
                    pygame.draw.circle(
                        canvas,
                        CI,
                        center,
                        pix_square_size / 2.5,
                    )
                    pygame.draw.circle(
                        canvas,
                        BG,
                        center,
                        pix_square_size / 2.5 - LW,
                    )

        return canvas

    def _get_img(self):
        canvas = self._get_frame()
        return np.transpose(
                np.array(pygame.surfarray.pixels3d(canvas)), axes=(1, 0, 2)
            ) 

    def close(self):
        if self.window is not None:
            pygame.display.quit()
            pygame.quit()


import gym
from gym import spaces
from . import adversarial

import chess
import chess.svg

import numpy as np

from io import BytesIO
import cairosvg
from PIL import Image
import pygame


class ChessActionSpace(adversarial.AdversarialActionSpace):
    def __init__(self, board):
        self.board = board

    @property
    def legal_actions(self):
        return [ChessEnv.move_to_action(move) for move in self.board.legal_moves]
    
    @property
    def action_space_size(self):
        return 64 * 73


class ChessEnv(adversarial.AdversarialEnv):
    """Chess Environment"""
    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 4}

    def __init__(self, render_mode=None, render_size=512, claim_draw=True, **kwargs):
        self.board = chess.Board(chess960=False)

        self.action_space = ChessActionSpace(self.board)
        self.observation_space = spaces.Tuple(spaces=(
            spaces.Box(low=-6, high=6, shape=(8, 8), dtype=np.int8),
            spaces.Box(low=np.array([False]),
                       high=np.array([True]), dtype=np.bool)
        ))

        self.render_size = render_size
        self.claim_draw = claim_draw

        assert render_mode is None or render_mode in self.metadata["render_modes"]
        self.render_mode = render_mode

        self.clock = None
        self.window = None

    @property
    def current_player(self):
        return self.board.turn

    @property
    def previous_player(self):
        return not self.board.turn

    @property
    def starting_player(self):
        return chess.WHITE

    def get_string_representation(self):
        return self.board.fen()

    def set_string_representation(self, board_string):
        self.board = chess.Board(board_string)
        self.action_space = ChessActionSpace(self.board)
        return self._get_canonical_observation()

    def _get_canonical_observation(self):
        state = (self.get_piece_configuration(self.board))
        player = self.current_player

        # canonical_representation = - \
        #     state[::-1, ::-1] if player == chess.BLACK else state

        canonical_representation = -state if player == chess.BLACK else state
        return canonical_representation, np.array([player], dtype=np.bool)

    def _get_info(self):
        info = {
            'castling_rights': self.board.castling_rights,
            'fullmove_number': self.board.fullmove_number,
            'halfmove_clock': self.board.halfmove_clock,
            'promoted': self.board.promoted,
            'ep_square': self.board.ep_square
        }
        return info

    def _do_action(self, action):
        move = self.action_to_move(action)
        self.board.push(move)
    
    def _reset_game(self):
        self.board.reset()

    def _game_result(self):
        result = self.board.result()
        winner = (chess.WHITE if result == '1-0' else chess.BLACK if result ==
                '0-1' else -1 if result == '1/2-1/2' else None)
        # result is 1 for white win or 0 for black win. slight positive for draw
        reward = 0 if result is None else 1e-4 if result == -1 else 1
        return winner, reward

    def _get_frame(self):
        surf = pygame.surfarray.make_surface(self._get_img())
        surf = pygame.transform.flip(surf, False, True)
        return pygame.transform.rotate(surf, -90)

    def _get_img(self):
        out = BytesIO()
        bytestring = chess.svg.board(
            self.board, size=self.render_size).encode('utf-8')
        cairosvg.svg2png(bytestring=bytestring, write_to=out)
        image = Image.open(out)
        img = np.asarray(image)
        return img   

    def close(self):
        if self.window is not None:
            pygame.display.quit()
            pygame.quit()

    def set_board_state(self, canonicial_state):
        canonicial_representation, player = canonicial_state
        state = -canonicial_representation[::-1, ::-
                                           1] if player == chess.BLACK else canonicial_representation

        piece_map = {}
        for square, piece in enumerate(state.flatten()):
            if piece:
                color = chess.Color(int(np.sign(piece) > 0))
                piece_map[chess.Square(square)] = chess.Piece(
                    chess.PieceType(abs(piece)), color)

        self.board.set_piece_map(piece_map)

    def action_to_move(self, action):
        unraveled_action = np.unravel_index(action, (64, 73))
        from_square = unraveled_action[0]

        if unraveled_action[1] < 64:
            to_square = unraveled_action[1]
            move = self.board.find_move(from_square, to_square)
        else:
            pd = unraveled_action[1] - 64
            unraveled_pd = np.unravel_index(pd, (3, 3))
            promotion = unraveled_pd[0] + 2

            from_file = chess.square_file(from_square)
            to_file = unraveled_pd[1] - 1 + from_file
            from_rank = chess.square_rank(from_square)
            to_rank = 0 if 1 == from_rank else 7
            to_square = chess.square(to_file, to_rank)
            move = self.board.find_move(from_square, to_square, promotion=promotion)
        return move

    @staticmethod
    def move_to_action(move):
        from_square = move.from_square
        to_square = move.to_square
        promotion = (0 if move.promotion is None else move.promotion)

        from_file = chess.square_file(from_square)
        to_file   = chess.square_file(to_square)

        if promotion == 0 or promotion == chess.QUEEN:
            action = (from_square, to_square)
            return np.ravel_multi_index(action, (64, 73))
        else:
            d = to_file - from_file + 1 # in {0, 1, 2}
            p = promotion - 2 # in {0, 1, 2}
            pd = np.ravel_multi_index((p, d), (3, 3))
            action = (from_square, 64 + pd) 
            return np.ravel_multi_index(action, (64, 73))

    @staticmethod
    def get_piece_configuration(board):
        piece_map = np.zeros(64, dtype=np.int8)

        for square, piece in board.piece_map().items():
            piece_map[square] = piece.piece_type * (2 * piece.color - 1)

        return piece_map.reshape((8, 8))


import gym
from gym import spaces

import numpy as np
import pygame
from abc import ABC, abstractmethod, abstractproperty

class AdversarialActionSpace(gym.spaces.Space):

    def sample(self):
        actions = self.legal_actions
        return actions[np.random.randint(len(actions))]

    def contains(self, action, is_legal=True):
        is_contained = action in range(self.action_space_size())
        and_legal = action in self.legal_actions if is_legal else True
        return is_contained and and_legal

    @abstractproperty
    def legal_actions(self):
        """
        Returns:
            legal_actions: Returns a list of all the legal moves in the current position.
        """
        pass
    
    @abstractproperty
    def action_space_size(self):
        """
        Returns:
            action_space_size: returns the number of all possible actions.
        """
        pass
    

  
class AdversarialEnv(gym.Env):
    """Abstract Adversarial Environment"""

    skip_next_render = False

    @abstractproperty
    def current_player(self):
        """
        Returns:
            current_player: Returns identifier for which player currently has their turn.
        """
        pass

    @abstractproperty
    def previous_player(self):
        """
        Returns:
            previous_player: Returns identifier for which player previously had their turn.
        """
        pass

    @abstractproperty
    def starting_player(self):
        """
        Returns:
            starting_player: Returns identifier for which player started the game.
        """
        pass

    @abstractmethod
    def get_string_representation(self):
        """
        Returns:
            board_string: Returns string representation of current game state.
        """
        pass
    
    @abstractmethod
    def set_string_representation(self, board_string):
        """
        Input:
            board_string: sets game state to match the string representation of board_string.
        Returns:
            canonical_state: returns canonical form of board. The canonical form
                            should be independent of players turn. For e.g. in chess,
                            the canonical form can be chosen to be from the pov
                            of white. When the player is white, we can return
                            board as is. When the player is black, we can invert
                            the colors and return the board.
            current_player: returns indentifier of which player is the current player in the canonicial state. 
                            This is used to decode the invariant canonical form.
        """
        pass

    @abstractmethod
    def _get_canonical_observation(self):
        """
        Returns:
            canonical_state: returns canonical form of board. The canonical form
                            should be independent of players turn. For e.g. in chess,
                            the canonical form can be chosen to be from the pov
                            of white. When the player is white, we can return
                            board as is. When the player is black, we can invert
                            the colors and return the board.
            current_player: returns indentifier of which player is the current player in the canonicial state. 
                            This is used to decode the invariant canonical form.
        """
        pass  

    @abstractmethod
    def _game_result(self):
        """
        Returns:
            winner: returns None when game is not finished else returns int value 
                    for the winning player or draw.
            reward: Reward value given the game result. Should not consider the player who won.
               
        """
        pass

    @abstractmethod
    def _do_action(self, action):
        """
        Input:
            action: Execute action from current game state.
        """
        pass
    
    @abstractmethod
    def _reset_game(self):
        """
        Reset the state of the game to the initial state. 
        This includes reseting the current player to the starting player.
        """

    @abstractmethod
    def _get_frame(self):
        """
        Returns:
            frame: returns py_game frame for the current state of the game. 
            This will be used by render to render the frame for human visualization
               
        """
        pass

    @abstractmethod
    def _get_img(self):
        """
        Returns:
            img: returns rgb_array of the image for the current state of the game. 
               
        """
        pass

    def game_result(self):
        return self._game_result()[0]

    def skip_next_human_render(self):
        """
        Skips the next automatic human render in step or reset. 
        Used for rollouts or similar non visualized moves.
        """
        self.skip_next_render = True

    def step(self, action):
        self._do_action(action)
        observation = self._get_canonical_observation()
        info = self._get_info()
        result, reward = self._game_result()
        terminated = result is not None

        if self.render_mode == "human":
            self.render()

        return observation, reward, terminated, False, info

    def reset(self, seed=None, options=None):
        super().reset(seed=seed)

        self._reset_game()
        observation = self._get_canonical_observation()
        info = self._get_info()

        if self.render_mode == "human":
            self.render()
        
        return observation, info
        
    def render(self):

        if self.render_mode == "human" and not self.skip_next_render:
            if self.clock is None:
                self.clock = pygame.time.Clock()
            if self.window is None:
                pygame.init()
                pygame.display.init()
                self.window = pygame.display.set_mode((self.render_size, self.render_size))

            canvas = self._get_frame()
            # The following line copies our drawings from `canvas` to the visible window
            self.window.blit(canvas, canvas.get_rect())
            pygame.display.update()
            # We need to ensure that human-rendering occurs at the predefined framerate.
            # The following line will automatically add a delay to keep the framerate stable.
            self.clock.tick(self.metadata["render_fps"])

        elif self.render_mode == "rgb_array":
            return self._get_img()

        self.skip_next_render = False



import gym
import adversarial_gym

# env = gym.make("Chess-v0", render_mode='human')
env = gym.make("TicTacToe-v0", render_mode='human')
print('reset')
env.reset()
terminal = False
while not terminal:
    action = env.action_space.sample()
    observation, reward, terminal, truncated, info = env.step(action)
env.close()


from setuptools import setup, find_packages

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='adversarial_gym',
      version='0.0.3',
      description='OpenAI Gym environments for adversarial games for the operation beat ourselves organisation.',
      url='https://github.com/OperationBeatMeChess/adversarial-gym',
      author='Dawson Horvath',
      author_email='horvath.dawson@gmail.com',
      license='MIT License',
      install_requires=['gym>=0.26.1', 'python-chess', 'numpy', 'cairosvg', 'pillow', 'pygame'],
      long_description=long_description,
      long_description_content_type="text/markdown",
)



from gym.envs.registration import register

register(
    id='Chess-v0',
    entry_point='adversarial_gym.chess_env:ChessEnv'
)
register(
    id='TicTacToe-v0',
    entry_point='adversarial_gym.tictactoe_env:TicTacToeEnv'
)